#include <strings.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>

#include "core.h"
#include "threads.h"
#include "sock.h"

#include "ocsp_request.h"
#include "ocsp_http.h"

extern OCSPD_CONFIG *ocspd_conf;

http_method_t
decode_http_method(const char *str, int len, int *error)
{
    http_method_t   method;

    switch (*str) {
    case 'p':
    case 'P':
        if (len < 5) {
            *error = HTTP_HEADER_TRUNCATED;
            return METHOD_ERROR;
        }

        if (strncasecmp(str, "POST ", 5)) {
            *error = HTTP_INVALID_METHOD;
            return METHOD_ERROR;
        }

        method = METHOD_POST;
        break;
    case 'g':
    case 'G':
        if (len < 4) {
            *error = HTTP_HEADER_TRUNCATED;
            return METHOD_ERROR;
        }

        if (strncasecmp(str, "GET ", 4)) {
            *error = HTTP_INVALID_METHOD;
            return METHOD_ERROR;
        }

        method = METHOD_GET;
        break;
    default:
        *error = HTTP_INVALID_METHOD;
        return METHOD_ERROR;
    }

    return method;
}

char           *
http_expand_hex(const char *hexstr)
{
    switch (strtol(hexstr, NULL, 16)) {
    case 0x24:
        return "$";
    case 0x26:
        return "&";
    case 0x2B:
        return "+";
    case 0x2C:
        return ",";
    case 0x2F:
        return "/";
    case 0x3A:
        return ":";
    case 0x3B:
        return ";";
    case 0x3D:
        return "=";
    case 0x3F:
        return "?";
    case 0x40:
        return "@";
    case 0x20:
        return " ";
    case 0x22:
        return "\"";
    case 0x3C:
        return "<";
    case 0x3E:
        return ">";
    case 0x23:
        return "#";
    case 0x25:
        return "%";
    case 0x7b:
        return "{";
    case 0x7d:
        return "}";
    case 0x7c:
        return "|";
    case 0x5c:
        return "\\";
    case 0x5e:
        return "^";
    case 0x7e:
        return "~";
    case 0x5B:
        return "[";
    case 0x5D:
        return "]";
    case 0x60:
        return "`";
    default:
        return NULL;
    }
    return NULL;
}

char           *
http_decode_base64(char *input, int len, int *tlen)
{
    BIO            *bio,
                   *b64;
    char           *outbuf;
    int             inlen;

    if (!(outbuf = malloc(4096))) {
        *tlen = -1;
        return NULL;
    }

    memset(outbuf, 0, 4096);

    b64 = BIO_new(BIO_f_base64());
    bio = BIO_new_mem_buf(input, len);
    bio = BIO_push(b64, bio);

    while ((inlen = BIO_read(bio, outbuf, 4096)) > 0)
        *tlen += inlen;

    BIO_free_all(bio);

    return outbuf;
}

char           *
decode_uri(char *in_uri, int len)
{
    int             i;
    int             current_byte;
    int             characters_written;
    int             outstring_len;
    char           *outstring;

    if (!in_uri || len <= 0)
        return NULL;

    /*
     * (len / 64) is the number of newline characters that will need to
     * be added.
     * 
     * We also add 2 more bytes to the end for the last \n, and a \0
     * terminator. 
     */

    outstring_len = (len / 64) + len + 2;
    characters_written = 0;

    if (!(outstring = malloc(outstring_len)))
        return NULL;

    memset(outstring, 0, outstring_len);

    current_byte = 0;

    for (i = 0; i < len; i++) {
        char           *data;
        char           *hexstr;
        char           *expanded;

        data = &in_uri[i];

        /*
         * openssl base64 decoding requires our encoded string to include 
         * newlines after every 64 bytes 
         */
        if (i > 0 && !(characters_written % 63)) {
            outstring[current_byte++] = '\n';
        }

        if (*data != '%') {
            outstring[current_byte++] = *data;
            characters_written++;
            continue;
        }

        if (i + 2 >= outstring_len) {
            /*
             * we can't actually fit a hex set here 
             */
            outstring[current_byte++] = *data;
            characters_written++;
            continue;
        }

        hexstr = malloc(3);
        hexstr[0] = data[1];
        hexstr[1] = data[2];
        hexstr[2] = '\0';

        expanded = http_expand_hex(hexstr);

        if (expanded) {
            outstring[current_byte++] = *expanded;
            i += 2;
        } else
            outstring[current_byte++] = *data;

        characters_written++;
        free(hexstr);
    }

    outstring[current_byte] = '\n';

    return outstring;
}

http_rec_t     *
create_http_rec(void)
{
    http_rec_t     *rec;

    if (!(rec = malloc(sizeof(http_rec_t))))
        return NULL;

    memset(rec, 0, sizeof(http_rec_t));
    return rec;
}

void
free_http_rec(http_rec_t * rec)
{
    if (!rec)
        return;

    if (rec->http_raw_rec)
        free(rec->http_raw_rec);

    if (rec->http_data)
        free(rec->http_data);

    free(rec);
}

int
decode_http_content_len(char *input_buf, int input_len)
{
    /*
     * return values: -1 error occurred 0 not enough data >0 the length to 
     * read 
     */
    char           *content_sz_raw;
    char           *content_sz_val_raw;
    char           *content_sz_val;
    char           *content_sz_end;
    int             content_sz_val_len;
    int             content_len;

    /*
     * first find the content-length header 
     */
    content_sz_raw = strstr(input_buf, "Content-Length: ");

    if (!content_sz_raw)
        return 0;

    /*
     * paranoid overflow check 
     */
    if (content_sz_raw > (char *) (input_buf + input_len))
        return 0;

    content_sz_val_raw = &content_sz_raw[16];

    if (!content_sz_val_raw)
        return 0;

    if (content_sz_val_raw > (char *) (input_buf + input_len))
        return 0;

    /*
     * find the end of the content-length line 
     */
    content_sz_end = strstr(content_sz_val_raw, "\r\n");

    if (!content_sz_end)
        return 0;

    /*
     * get a general number of bytes to alloc 
     */
    content_sz_val_len = content_sz_end - content_sz_val_raw;

    if (content_sz_val_len <= 0)
        return -1;

    if (!(content_sz_val = malloc(content_sz_val_len)))
        return -1;

    memset(content_sz_val, 0, content_sz_val_len + 1);
    strncpy(content_sz_val, content_sz_val_raw, content_sz_val_len);

    content_len = atoi(content_sz_val);

    free(content_sz_val);

    if (content_len >= INT_LEAST32_MAX)
        return -1;

    return content_len;
}

char           *
decode_http_post_data(char *input_buf, int input_len,
                      int content_len, int *error)
{
    char           *headers_end;
    char           *payload_start;
    char           *input_buf_end;
    int             current_content_len;


    input_buf_end = (char *) (input_buf + input_len);

    /*
     * find the end of all the headers 
     */
    headers_end = strstr(input_buf, "\r\n\r\n");

    if (!headers_end)
        headers_end = strstr(input_buf, "\n\n");

    if (!headers_end) {
        /*
         * not enough data? continue on 
         */
        *error = HTTP_CONTINUE;
        return NULL;
    }

    payload_start = headers_end;

    while ((*payload_start == '\n') || (*payload_start == '\r')) {
        if (payload_start > input_buf_end)
            break;

        payload_start++;
    }

    if (payload_start > input_buf_end) {
        *error = HTTP_OVERFLOW;
        return NULL;
    }

    /*
     * check to see if we have the entire payload as determined by the
     * content_len 
     */
    current_content_len = input_buf_end - payload_start;

    if (current_content_len < content_len) {
        *error = HTTP_CONTINUE;
        return NULL;
    }

    if (current_content_len > content_len) {
        *error = HTTP_OVERFLOW;
        return NULL;
    }

    return payload_start;
}


http_rec_t     *
decode_http_post(char *input_buf, int input_len, int *error)
{
    int             content_len;
    char           *post_data;
    http_rec_t     *rec;

    content_len = decode_http_content_len(input_buf, input_len);

    if (content_len == 0) {
        *error = HTTP_CONTINUE;
        return NULL;
    }

    if (content_len < 0) {
        *error = HTTP_INVALID_CONTENTLEN;
        return NULL;
    }

    post_data = decode_http_post_data(input_buf,
                                      input_len, content_len, error);

    if (!post_data)
        return NULL;

    if (!(rec = create_http_rec())) {
        *error = HTTP_ALLOC_ERROR;
        return NULL;
    }

    rec->method = METHOD_POST;
    rec->request_len = input_len;
    rec->content_len = content_len;

    rec->http_raw_rec = malloc(input_len);
    memcpy(rec->http_raw_rec, input_buf, input_len);

    rec->http_data = malloc(content_len);
    memcpy(rec->http_data, post_data, content_len);

    return rec;
}

http_rec_t     *
decode_http_get(char *input_buf, int input_len, int *error)
{
    /*
     * decoding a ocsp GET request is a trivial task. Simply read the
     * URI, hop past the base_uri, swap reserved characters with the real 
     * values, and base64 decode the string 
     */
    http_rec_t     *rec;
    char           *uri_start;
    char           *uri_end;
    char           *uri;
    char           *decoded_uri;
    char           *unbased_uri;
    int             uri_len;
    int             tlen;

    rec = NULL;
    tlen = 0;

    /*
     * lets first find the base URI 
     */
    uri_start = strstr(input_buf, ocspd_conf->base_uri);

    if (!uri_start) {
        *error = HTTP_BASEURI_ERROR;
        return NULL;
    }

    if (uri_start > (input_buf + input_len)) {
        *error = HTTP_OVERFLOW;
        return NULL;
    }

    /*
     * increment past the base_uri 
     */
    uri_start += strlen(ocspd_conf->base_uri);

    /*
     * find the end of the GET request by looking for the protocol 
     */
    uri_end = strstr(uri_start, " HTTP/1");

    if (!uri_end) {
        *error = HTTP_HEADER_TRUNCATED;
        return NULL;
    }

    if (uri_end > (input_buf + input_len)) {
        *error = HTTP_OVERFLOW;
        return NULL;
    }

    uri_len = uri_end - uri_start;

    if (!(uri = malloc(uri_len + 1))) {
        *error = HTTP_ALLOC_ERROR;
        return NULL;
    }

    memset(uri, 0, uri_len + 1);
    memcpy(uri, uri_start, uri_len);

    /*
     * do our reserved hex encoded swapping with the real thing 
     */
    decoded_uri = decode_uri(uri, uri_len);
    free(uri);

    if (!decoded_uri) {
        *error = HTTP_URI_DECODE;
        return NULL;
    }

    /*
     * now take the input uri and base64 decode 
     */
    unbased_uri = http_decode_base64(decoded_uri,
                                     strlen(decoded_uri), &tlen);

    if (!unbased_uri || tlen <= 0) {
        *error = HTTP_BASE64_DECODE;

#if 0
        printf("DECODED %p tlen %d\n", unbased_uri, tlen);

        printf("IN BASE64\n=================\n%s\n=================\n",
               decoded_uri);
#endif

        if (unbased_uri)
            free(unbased_uri);

        if (decoded_uri)
            free(decoded_uri);

        return NULL;
    }

    if (!(rec = create_http_rec())) {
        if (decoded_uri)
            free(decoded_uri);

        if (unbased_uri)
            free(unbased_uri);

        *error = HTTP_HEADER_TRUNCATED;
        return NULL;
    }

    rec->method = METHOD_GET;
    rec->request_len = input_len;
    rec->content_len = tlen;
    rec->http_raw_rec = strdup(input_buf);
    rec->http_data = unbased_uri;

    if (decoded_uri)
        free(decoded_uri);

    return rec;
}

http_rec_t     *
decode_http_request(char *input_buf, int input_len, int *error)
{
    http_method_t   method;
    http_rec_t     *rec;

    rec = NULL;

    method = decode_http_method(input_buf, input_len, error);

    switch (method) {
    case METHOD_POST:
        rec = decode_http_post(input_buf, input_len, error);
        break;
    case METHOD_GET:
        rec = decode_http_get(input_buf, input_len, error);
        break;
    default:
        return NULL;
    }

    return rec;
}
